try:
    from xml.etree import ElementTree
except ImportError:
    from elementtree import ElementTree

from urllib.parse import urlencode
from urllib.request import urlopen
from urllib.parse import urljoin

import ocflib.account.search as search
import ocflib.constants as constants


def verify_ticket(ticket, service):
    """Verifies CAS 2.0+ XML-based authentication ticket.

    Returns CalNet UID on success and None on failure.
    """
    params = {'ticket': ticket, 'service': service}
    url = (urljoin(constants.CAS_URL, 'serviceValidate') + '?' +
           urlencode(params))
    try:
        page = urlopen(url)
        response = page.read()
        tree = ElementTree.fromstring(response)
        if tree[0].tag.endswith('authenticationSuccess'):
            return tree[0][0].text
        else:
            return None
    except:
        return None


def get_calnet_names(uid):
    """Returns CalNet LDAP entries relating to names"""

    attrs = search.user_attrs_ucb(uid)
    if attrs:
        return {key: attrs[key]
                for key in ('givenName', 'sn', 'displayname') if attrs[key]}


def name_by_calnet_uid(uid):
    """Returns the name of CalNet person, searched by CalNet UID.

    Returns None on faliure.
    """
    names = get_calnet_names(uid)

    # the name we want to input into our system is "givenName sn"
    # displayName is not necessarily equal to what's printed on Cal 1 Cards
    get_longest_string = lambda strs: max(strs, key=len)

    if 'givenName' in names and 'sn' in names:
        given_name = get_longest_string(names['givenName'])
        sn = get_longest_string(names['sn'])

        if given_name and sn:
            return "{} {}".format(given_name, sn)
    elif 'displayName' in names:
        display_name = get_longest_string(names["displayName"])

        if display_name:
            return display_name



"""Random account methods that don't fit anywhere else."""
import pexpect
import requests

import ocflib.account.validators as validators
import ocflib.constants as constants


def password_matches(username, password):
    """Returns True if the password matches the user account given"""

    validators.validate_username(username)
    validators.validate_password(username, password, strength_check=False)

    if not validators.user_exists(username):
        raise ValueError("User doesn't exist")

    cmd = "kinit --no-forwardable -l0 {}@OCF.BERKELEY.EDU".format(username)
    child = pexpect.spawn(cmd, timeout=10)

    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(username))
    child.sendline(password)

    child.expect(pexpect.EOF)
    child.close()

    return child.exitstatus == 0


def has_vhost(user):
    """Returns whether or not a virtual host is already configured for
    the given user."""

    check = (user, user + "!")
    line_matches = lambda fields: len(fields) > 0 and fields[0] in check

    vhosts = requests.get(constants.VHOST_DB_URL).text.split("\n")
    return any(line_matches(line.split()) for line in vhosts)



import difflib
import pwd
import string

import cracklib

import ocflib.constants as constants


def validate_username(username):
    """Validate a username, raising a descriptive exception if problems are
    encountered."""

    if username_reserved(username):
        raise ValueError("Username is reserved")

    if not 3 <= len(username) <= 8:
        raise ValueError("Username must be between 3 and 8 characters")

    if not all(c.islower() for c in username):
        raise ValueError("Username must be all lowercase letters")


def validate_password(username, password, strength_check=True):
    """Validate a password, raising a descriptive exception if problems are
    encountered. Optionally checks password strength."""

    if strength_check:
        if len(password) < 8:
            raise ValueError("Password must be at least 8 characters")

        s = difflib.SequenceMatcher()
        s.set_seqs(password, username)

        if s.ratio() > 0.6:
            raise ValueError("Password is too similar to username")

        try:
            cracklib.FascistCheck(password)
        except ValueError as e:
            raise ValueError("Password problem: {}".format(e))

    # sanity check; note we don't use string.whitespace since we don't want
    # tabs or newlines
    allowed = string.digits + string.ascii_letters + string.punctuation + ' '

    if not all(c in allowed for c in password):
        raise ValueError("Password contains forbidden characters")


def user_exists(username):
    try:
        pwd.getpwnam(username)
    except KeyError:
        return False
    else:
        return True


def username_reserved(username):
    if username.startswith('ocf'):
        return True
    return username in constants.RESERVED_USERNAMES


def username_queued(username):
    """Returns if the username has already been requested and is queued to be
    created."""
    with open(constants.QUEUED_ACCOUNTS_PATH) as f:
        return any(line.startswith(username + ":") for line in f)
